/*
	This file contains specific functions depending on the dll-package
	that deal with regular expressions and ACLs.
*/ 

#include <string.h>
#include <stdlib.h>
#include <unistd.h>

#include "types.h"
#include "regex.h"
#include "globals.h"
#include "access.h"
#include "dll.h"
#include "dllapps.h"


in_addr_t xinet_addr (CHAR *host, INT resolv)
{
    in_addr_t addr;

    addr =  ((in_addr_t)inet_addr(host));

    if ((addr == (in_addr_t)INADDR_NONE) && resolv) {
        struct hostent *hp;
        hp = gethostbyname(host);
        if (hp != NULL)
            addr = *((in_addr_t *)hp->h_addr);
    }

    return addr;
}


static VOID getHeadTail (VOID *ehead, VOID *headp, VOID *tailp)
{
    dll_t *p;

    p = (dll_t *)ehead;
    if (headp != NULL) *((dll_t **)headp) = p;
    if (tailp != NULL) *((dll_t **)tailp) = p+1;
}


VOID clearReChain (dlr_t *head)
{
    clearChain ((dll_t *)head);
}

VOID clearPidChain (dlp_t *head)
{
    clearChain ((dll_t *)head);
}

VOID clearFilterChain (dlf_t *head)
{
    clearChain ((dll_t *)head);
}

VOID deleteReChain (dlr_t *head)
{
    dlr_t *ptr, *tail; 

    getHeadTail (head, NULL, &tail);

    for (ptr = head->next; ptr != tail; ptr = ptr->next) {
        regfree(&ptr->compiled);
        if (ptr->raw != NULL)
            free (ptr->raw);
        if (ptr->map != NULL)
            free (ptr->map);
    }

    deleteChain((dll_t *)head);
}


VOID deleteACLChain (dla_t *head)
{
    deleteChain((dll_t *)head);
}


VOID deletePidChain (dlp_t *head)
{
    deleteChain((dll_t *)head);
}

VOID deleteFilterChain (dlf_t *head)
{
    dlf_t *ptr, *tail; 

    getHeadTail (head, NULL, &tail);

    for (ptr = head->next; ptr != tail; ptr = ptr->next) {
        if (ptr->rxfilter != NULL)
            free (ptr->rxfilter);
        if (ptr->txfilter != NULL)
            free (ptr->txfilter);
    }

    deleteChain((dll_t *)head);
}



dlr_t *initReChain(VOID)
{
    return ((dlr_t *)initChain(NULL));
}


dla_t *initACLChain(VOID)
{
    return ((dla_t *)initChain(NULL));
}


dlp_t *initPidChain(VOID)
{
    return ((dlp_t *)initChain(NULL));
}

dlf_t *initFilterChain(VOID)
{
    return ((dlf_t *)initChain(NULL));
}


VOID deleteACLByMark (dla_t *head, INT mark)
{
    deleteByMark ((dll_t *)head, mark);
}

VOID deleteFilterByMark (dlf_t *head, INT mark)
{
    deleteByMark ((dll_t *)head, mark);
}


VOID deleteReByMark (dlr_t *head, INT mark)
{
    dlr_t *ptr, *tail;

    getHeadTail (head, NULL, &tail);

    for (ptr = head->next; ptr != tail; ) {
        if (ptr->mark == mark) {
            regfree(&ptr->compiled);
            if (ptr->raw != NULL)
                free (ptr->raw);
            if (ptr->map != NULL)
                free (ptr->map);
            ptr = (dlr_t *)removeElement((dll_t *)ptr, DLL_GETNEXT|DLL_FREE);
        }
        else
            ptr = ptr->next;
    }
}


INT deleteByPid (dlp_t *head, pid_t pid)
{
    dlp_t *ptr, *tail;
    INT rtc = 1;

    getHeadTail (head, NULL, &tail);

    for (ptr = head->next; ptr != tail; ) {
        if (ptr->pid == pid) {
            ptr = (dlp_t *)removeElement((dll_t *)ptr, DLL_GETNEXT|DLL_FREE);
            rtc = 0;
        }
        else
            ptr = ptr->next;
    }
    return rtc;
}


dlr_t *findRePattern (dlr_t *head, dlr_t *start, INT mark, CHAR *name)
{
    dlr_t *ptr, *tail;

    getHeadTail (head, NULL, &tail);

    if (start == NULL)
        start = head;

    for (ptr = start->next; ptr != tail; ptr = ptr->next) {
        if (mark == 0 || ptr->mark == mark)
            if (!regexec(&ptr->compiled, name, 0, NULL, 0))
                return ptr;
    }
    return NULL;	
}

dla_t *findACL (dla_t *head, dla_t *start, INT mark, in_addr_t addr)
{
    dla_t *ptr, *tail;

    getHeadTail (head, NULL, &tail);

    if (start == NULL)
        start = head;

    for (ptr = start->next; ptr != tail; ptr = ptr->next)
        if (mark == 0 || ptr->mark == mark) {
            if ((addr & ptr->acl.mask) == (ptr->acl.addr & ptr->acl.mask))
                return ptr;
        }
    return NULL;
}


dlp_t *findPid (dlp_t *head, pid_t pid)
{
    dlp_t *ptr, *tail;

    getHeadTail (head, NULL, &tail);

    for (ptr = head->next; ptr != tail; ptr = ptr->next)
        if (ptr->pid == pid)
            return ptr;

    return NULL;
}


dlr_t *getNextRe (dlr_t *head, dlr_t *start)
{
    return((dlr_t *)getNext((dll_t *)head, (dll_t *)start));
}


dla_t *getNextACL (dla_t *head, dla_t *start)
{
    return((dla_t *)getNext((dll_t *)head, (dll_t *)start));
}


dlf_t *getNextFilter (dlf_t *head, dlf_t *start)
{
    return((dlf_t *)getNext((dll_t *)head, (dll_t *)start));
}


dlp_t *getNextPid (dlp_t *head, dlp_t *start)
{
    return((dlp_t *)getNext((dll_t *)head, (dll_t *)start));
}


INT getPidCount (dlp_t *head)
{
    dll_t *dp;
    INT count;

    for (dp = NULL, count = 0; (dp = getNext((dll_t *)head, dp)) != NULL;)
        count++;

    return count;
}


dla_t *getACLByMark (dla_t *head, dla_t *start, INT mark)
{
    return ((dla_t *)getByMark ((dll_t *)head, (dll_t *)start, mark));
}


dlf_t *getFilterByMark (dlf_t *head, dlf_t *start, INT mark)
{
    return ((dlf_t *)getByMark ((dll_t *)head, (dll_t *)start, mark));
}

dlr_t *getReByMark (dlr_t *head, dlr_t *start, INT mark)
{
    return ((dlr_t *)getByMark ((dll_t *)head, (dll_t *)start, mark));
}

INT addRePattern (dlr_t *head, INT mark, CHAR *pattern, CHAR *map, INT filter, dlr_t **np)
{
    dlr_t *new;
    INT rtc;

    if ((new = (dlr_t *)malloc(sizeof(dlr_t))) == NULL)
        return -1;

    if ((new->raw = (CHAR *)malloc(strlen(pattern)+1)) == NULL) {
        free (new);
        return -1;
    }
    else {
        strcpy (new->raw, pattern);
    }

    if (map != NULL) {
        if ((new->map = (CHAR *)malloc(strlen(map)+1)) == NULL) {
            free (new->raw);
            free (new);
            return -1;
        }
        else {
            strcpy (new->map, map);
        }
    }
    else {
        new->map = NULL;
    }
    

    strcpy (new->raw, pattern);

    new->ctype  = CTYPE_REGEX;
    new->mark   = mark;
    new->filter = filter;

    if ((rtc = regcomp(&new->compiled, pattern, 0))) {
        free (new->raw);
        free (new->map);
        free (new);
        return rtc;
    }

    insertAtTail ((dll_t *)head, (dll_t *)new);
    if (np != NULL) *np = new;

    return 0;
}


INT addACL (dla_t *head, INT mark, INT perm, CHAR *addr, CHAR *mask, dla_t **np)
{
    dla_t *dp, *new, *tail;

    getHeadTail (head, NULL, &tail);

    if ((new = (dla_t *)malloc(sizeof(dla_t))) == NULL)
        return -1;

    new->ctype = CTYPE_ACL;
    new->mark = mark;
    new->acl.perm = perm;

    if (!strcasecmp(addr, "host")) {
        new->acl.mask = IPMASK_HOST;
        new->acl.addr = xinet_addr(mask, (!noResolv));
        if (new->acl.addr == INADDR_NONE) {
            free (new);
            return -2;
        }
    }
    else if (!strcasecmp(addr, "any")) {
        new->acl.mask = IPMASK_ANY;
        new->acl.addr = IPMASK_ANY;
    }
    else {
        new->acl.mask = ((in_addr_t)inet_addr(mask));
        new->acl.addr = xinet_addr(addr, (!noResolv));
        if (new->acl.addr == INADDR_NONE) {
            free (new);
            return -2;
        }
    }

    for (dp = head->next; dp != tail; dp = dp->next)
        if (dp->mark == mark) {
            if (ntohl(~dp->acl.mask) > ntohl(~new->acl.mask)) 
                break;
        }

    insertBefore((dll_t *)dp, (dll_t *)new);
    if (np != NULL) *np = new;

    return 0;
}

INT addPid (dlp_t *head, INT mark, pid_t pid, pid_t ppid, INT signal,
            uid_t uid, gid_t gid, dlp_t **np)
{
    dlp_t *new, *tail;

    getHeadTail (head, NULL, &tail);

    if ((new = (dlp_t *)malloc(sizeof(dlp_t))) == NULL)
        return -1;

    new->ctype  = CTYPE_PID;
    new->mark   = mark;
    new->pid    = pid;
    new->ppid   = ppid;
    new->signal = signal;
    new->uid    = uid;
    new->gid    = gid;

    insertBefore((dll_t *)tail, (dll_t *)new);
    if (np != NULL) *np = new;

    return 0;
}

INT addFilter (dlf_t *head, INT mark, CHAR *rxf, CHAR *txf, dlf_t **np)
{
    dlf_t *new, *tail;

    getHeadTail (head, NULL, &tail);

    if ((new = (dlf_t *)malloc(sizeof(dlf_t))) == NULL)
        return -1;

    if ((new->rxfilter = (CHAR *)malloc(strlen(rxf)+1)) == NULL) {
        free (new);
        return -1;
    }

    if ((new->txfilter = (CHAR *)malloc(strlen(txf)+1)) == NULL) {
        free (new->rxfilter);
        free (new);
        return -1;
    }

    new->ctype = CTYPE_FILTER;
    new->mark = mark;
    strcpy (new->rxfilter, rxf);
    strcpy (new->txfilter, txf);

    insertBefore((dll_t *)tail, (dll_t *)new);
    if (np != NULL) *np = new;

    return 0;
}

VOID sortFilterByMark (dlf_t *head, INT mode)
{
    sortByMark ((dll_t *)head, mode);
}

VOID modifyPLE (dlp_t *pp, pid_t pid, pid_t ppid, INT signal, 
                uid_t uid, gid_t gid)
{
    if (pid > 0)
        pp->pid = pid;
    else if (pid == 0)
        pp->pid = getpid();

    if (ppid > 0)
        pp->ppid = ppid;
    else if (pid == 0)
        pp->pid = getppid();

    if (uid > 0)
        pp->uid = uid;
    else if (uid == 0)
        pp->uid = geteuid();

    if (gid > 0)
        pp->gid = gid;
    else if (gid == 0)
        pp->gid = getegid();

    if (signal >= 0)
        pp->signal = signal;
}
